'''
Version and license information.
'''

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '0.1.5'
__versiondate__ = '2024-01-23'
__license__ = f'STIsim {__version__} ({__versiondate__}) — © 2024 by IDM'
